//---------------------------------------------------------------------------
/*
 * usb.c
 *
 * Linux driver for Demmel iLCD DPC3080 USB interface, based on libusb 1.0
 *
 * Written by Arlet Ottens <arlet@c-scape.nl>
 */

//#define DEBUG

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <assert.h>
#include <string.h>
#include <pty.h>
#include <utmp.h>
#include <poll.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <libusb-1.0/libusb.h>
#include "usb.h"

/*
 * The interface that we use. The iLCD only uses interface 0.
 */
#define INTERFACE   0

#define HID_REPORT_SET 0x09

/*
 * The vendor/product ID of the iLCD
 */
#define PRODUCT "iLCD Controller"
#define PRODUCT_LEN sizeof(PRODUCT)

/*
 * Size of the IRQ/CTRL buffer
 */
#define IRQ_SIZE 64
#define CTRL_SIZE 64

static int tx_timeout = 100;
static int rx_timeout = 100;

static libusb_context *context;

/*
 * USB device list
 */

static libusb_device_handle *lcd;

//---------------------------------------------------------------------------
/*
 * search the bus for Demmel iLCD device
 * return device handle if found, otherwise NULL.
 */
libusb_device_handle *search_device(void)
{
	libusb_device ** list = 0;
	ssize_t cnt = libusb_get_device_list(NULL, &list);
	ssize_t i = 0;
	unsigned char product[PRODUCT_LEN + 1];
	libusb_device_handle *handle;

	char buf[20];
	int saved_stderr = dup(2);

	for ( i = 0; i < cnt; i++ )
	{
		libusb_device *dev = list[i];
		struct libusb_device_descriptor desc;
		int res = libusb_get_device_descriptor(dev, &desc);
		FILE *fp;

		if (res < 0)
		{
			Error_SetErr(USB_ERR_ENUM_DEV_SET);
			return 0;
		}

        // set stderr to null to avoid displaying permission failures
	
		fp = freopen("/dev/null", "w", stderr);

		res = libusb_open(dev, &handle);

		sprintf(buf, "/dev/fd/%d", saved_stderr);
		fp = freopen(buf, "w", stderr);


		// Permission etc. failed -> try next device
		if (res)
			continue;

		res = libusb_get_string_descriptor_ascii(handle, desc.iProduct, product, sizeof(product));

		if (res > 0 && strcmp((char *) product, PRODUCT) == 0)
		{
		    libusb_free_device_list(list, 1);
			return handle;
		}

		libusb_close(handle);
	}

	libusb_free_device_list(list, 1);
	return 0;
}

//---------------------------------------------------------------------------
/*
 * USB_Open: open the USB device. Return true if successful
 */
bool USB_Open(void)
{
	if (libusb_init(&context))
	{
        Error_SetErr(USB_ERR_LIBUSB_INIT);
		return false;
	}

	libusb_set_debug(context, 0);
	lcd = search_device();

	if (!lcd)
	{
	    Error_SetErr(USB_ERR_NO_DEVICES_FOUND);
		return false;
	}

	if (libusb_kernel_driver_active(lcd, INTERFACE))
		libusb_detach_kernel_driver(lcd, INTERFACE);

	return libusb_claim_interface(lcd, INTERFACE) >= 0;
}

//---------------------------------------------------------------------------
/*
 * USB_Transmit: Transmit a block of data. The data is broken up into
 * control packets, with the first byte indicating the length
 */
bool USB_Transmit(byte *data, DWORD length)
{
	int burst;
	int err;
	unsigned char buf[CTRL_SIZE];

	while (length > 0)
	{
		burst = min(length, CTRL_SIZE - 1);
		buf[0] = burst;
		memcpy(buf + 1, data, burst);
		err =
			libusb_control_transfer(lcd, LIBUSB_ENDPOINT_OUT | LIBUSB_REQUEST_TYPE_CLASS | LIBUSB_RECIPIENT_INTERFACE,
				HID_REPORT_SET, 0x200, 0, buf, burst + 1, tx_timeout);

		if (err < 0)
		{
            Error_SetErr(USB_ERR_WRITE);
			return false;
		}

		data += burst;
		length -= burst;
	}
	return true;
}

//---------------------------------------------------------------------------
/*
 * USB_ReceiveChar: read a single character. If the buffer is empty,
 * read a new interrupt packet from the device
 */
bool USB_ReceiveChar(unsigned char *c)
{
	static unsigned char rx_buf[IRQ_SIZE];
	static int rx_length = 0;
	static int rx_ptr = 0;
	int length;

	if (rx_ptr == rx_length)
	{
		libusb_interrupt_transfer(lcd, 1 | LIBUSB_ENDPOINT_IN, rx_buf, IRQ_SIZE, &length, rx_timeout);

		if (length == 0)
		{
		    Error_SetErr(USB_ERR_RX_NO_DATA);
			return false;
		}
		rx_length = 1 + rx_buf[0];
		rx_ptr = 1;
	}

	if (rx_ptr == rx_length)
	{
        Error_SetErr(USB_ERR_RX_NO_DATA);
		return false;
	}

	*c = rx_buf[rx_ptr++];

	return true;
}

//---------------------------------------------------------------------------
/*
 * USB_SetTimeouts:
 * set the Receive and Transmit timeouts in milliseconds
 */
void USB_SetTimeouts(dword rx_tout, dword tx_tout)
{
	rx_timeout = rx_tout;
	tx_timeout = tx_tout;
}

//---------------------------------------------------------------------------
/*
 * USB_Close: close the device, and free resources
 */
bool USB_Close(void)
{
	libusb_close(lcd);
	libusb_exit(context);
	return true;
}
//---------------------------------------------------------------------------
